#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <stdint.h>

/**
 * 
 * Offset	Size	Type		Comment
 * 	0	4	DWORD		ID of YM5 format. ('YM5!')
 * 	4	8	string[8]	Check String ('LeOnArD!')
 * 	12	4	DWORD		Nb of valid VBL of the file.
 * 	16	4	DWORD		Song attributes (see bellow)
 * 	20	2	WORD		Nb of digi-drum sample (can be 0)
 * 	22	4	DWORD		YM2149 External frequency in Hz
 * 					(ex:2000000 for ATARI-ST version,
 * 					    1000000 for AMSTRAD CPC)
 * 	26	2	WORD		Player frequency in Hz
 * 					(Ex: 50Hz for almost player)
 * 	28	4	DWORD		Vbl number to loop the song.
 * 					(0 is default)
 * 	32	2	WORD		Size (in bytes) of futur
 * 					additinal data.
 * 					(must be 0 for the moment)
 * 
 * 	For each digidrum sample:
 * 	{
 * 	..	4	DWORD		sample size
 *         ..      ?       BYTES           sample data (8bits per sample)
 * 	}
 * 
 * 	..	?	NT-String	Name of the song.
 * 	..	?	NT-String	Name of the author.
 * 	..	?	NT-String	Comments (YM file converter ?!)
 * 	..	?			All YM2149 registers.
 * 	..	4	DWORD		End-File check. ('End!')
 * 
 */

uint32_t get_word(FILE *fp)
{
	uint8_t buff[2];
	if (fread(buff, 1, 2, fp) != 2)
		return -1;
	return ((uint32_t)buff[0] << 8) | buff[1];
}

int put_word(FILE *fp, uint32_t val)
{
	uint8_t buff[2];
	buff[0] = (uint8_t)(val >> 8);
	buff[1] = (uint8_t)val;
	if (fwrite(buff, 1, 2, fp) != 2)
		return -1;
	return 0;
}

uint32_t get_dword(FILE *fp)
{
	uint8_t buff[4];
	if (fread(buff, 1, 4, fp) != 4)
		return -1;
	return ((uint32_t)buff[0] << 24) | ((uint32_t)buff[1] << 16) | ((uint32_t)buff[2] << 8) | buff[3];
}

int put_dword(FILE *fp, uint32_t val)
{
	uint8_t buff[4];
	buff[0] = (uint8_t)(val >> 24);
	buff[1] = (uint8_t)(val >> 16);
	buff[2] = (uint8_t)(val >> 8);
	buff[3] = (uint8_t)val;
	if (fwrite(buff, 1, 4, fp) != 4)
		return -1;
	return 0;
}

int convert(const char *filename, FILE *fi, FILE *fo)
{
	uint8_t buff[32];
	uint8_t *regs[16];
	char song[512];
	char author[512];
	char software[512];
	char *dst;
	uint8_t **drums;
	uint32_t nvbl;
	uint32_t attr;
	uint32_t ndrums;
	uint32_t clock;
	uint32_t vbl;
	uint32_t loop;
	uint32_t extra;
	uint32_t n, m;


	if (4 != fread(buff, 1, 4, fi)) {
		fprintf(stderr, "EOF on input '%s'\n", filename);
		return -1;
	}
	buff[4] = '\0';
	if (strcmp(buff, "YM5!")) {
		fprintf(stderr, "'%s' not a YM5 file\n", filename);
		return -1;
	}
	if (8 != fread(buff, 1, 8, fi)) {
		fprintf(stderr, "EOF on input '%s'\n", filename);
	}
	buff[8] = '\0';
	if (strcmp(buff, "LeOnArD!")) {
		fprintf(stderr, "'%s' not a valid YM file\n", filename);
		return -1;
	}

	nvbl = get_dword(fi);
	attr = get_dword(fi);
	ndrums = get_word(fi);
	clock = get_dword(fi);
	vbl = get_word(fi);
	loop = get_dword(fi);
	extra = get_word(fi);
	fseek(fi, extra, SEEK_CUR);
	if (ndrums > 0) {
		drums = malloc(ndrums * sizeof(uint8_t *));
		if (!drums) {
			perror("malloc()");
			return -1;
		}
		for (n = 0; n < ndrums; n++) {
			uint32_t size = get_dword(fi);
			drums[n] = malloc(size + 4);
			if (!drums[n]) {
				perror("malloc()");
				return -1;
			}
			memcpy(&drums[n], &size, 4);
			if (size != fread(&drums[n][4], 1, size, fi)) {
				fprintf(stderr, "EOF on input '%s'\n", filename);
				return -1;
			}
		}
	}
	dst = song;
	while ((*dst++ = fgetc(fi)) && dst < &song[511])
		;
	dst = author;
	while ((*dst++ = fgetc(fi)) && dst < &author[511])
		;
	dst = software;
	while ((*dst++ = fgetc(fi)) && dst < &software[511])
		;
	for (n = 0; n < 16; n++) {
		regs[n] = malloc(nvbl);
		if (!regs[n]) {
			perror("malloc()");
			return -1;
		}
	}
	for (n = 0; n < 16; n++) {
		if (fread(regs[n], 1, nvbl, fi) != nvbl) {
			fprintf(stderr, "EOF in input '%s'\n", filename);
			return -1;
		}
	}
	if (fread(buff, 1, 4, fi) != 4) {
		fprintf(stderr, "EOF in input '%s'\n", filename);
		return -1;
	}
	buff[4] = '\0';
	if (strcmp(buff, "End!")) {
		fprintf(stderr, "\"End!\" marker not found in input '%s'\n", filename);
		return -1;
	}

	snprintf(buff, sizeof(buff), "YM6!");
	fwrite(buff, 1, 4, fo);
	snprintf(buff, sizeof(buff), "LeOnArD!");
	fwrite(buff, 1, 8, fo);
	put_dword(fo, nvbl);
	attr &= ~1;			/* reset bit 0 (not interleaved) */
	put_dword(fo, attr);
	put_word(fo, ndrums);
	put_dword(fo, clock);
	put_word(fo, vbl);
	put_dword(fo, loop);
	put_word(fo, 0);		/* no extra */
	if (ndrums > 0) {
		for (n = 0; n < ndrums; n++) {
			uint32_t size;
			memcpy(&size, &drums[n], 4);
			put_dword(fo, size);
			if (fwrite(&drums[n][4], 1, size, fo) != size) {
				fprintf(stderr, "File system full?\n");
				return -1;
			}
			free(drums[n]);
		}
		free(drums);
	}
	fwrite(song, 1, strlen(song) + 1, fo);
	fwrite(author, 1, strlen(author) + 1, fo);
	fwrite(software, 1, strlen(software) + 1, fo);
	for (n = 0; n < nvbl; n++) {
		for (m = 0; m < 16; m++)
			fwrite(&regs[m][n], 1, 1, fo);
	}
	snprintf(buff, sizeof(buff), "End!");
	fwrite(buff, 1, 4, fo);
	return 0;
}


int main(int argc, char **argv)
{
	char tempname[FILENAME_MAX];
	char *ext;
	FILE *fi, *fo;
	int i;

	for (i = 1; i < argc; i++) {
		snprintf(tempname, sizeof(tempname), "%s", argv[i]);
		ext = strrchr(tempname, '.');
		if (!ext)
			continue;
		snprintf(ext, sizeof(tempname) - (int)(ext - tempname), ".tmp");
		fi = fopen(argv[i], "rb");
		if (!fi) {
			perror(argv[i]);
			continue;
		}
		fo = fopen(tempname, "wb");
		convert(argv[i], fi, fo);
		fclose(fo);
		fclose(fi);
	}

	return 0;
}
